
const subTableId = "#subTable";

// #region Init Module

function initSubscriptionsModule() {
    initializeDataTable(
        "#subTable",
        [
            {
                title: "Name",
                data: "name",
                className: "text-truncate",
                render: function(data, type, row) {
                    const name = sanitise(data);
                    return `<button type="button" class="btn btn-link text-start text-decoration-none">${name}</button>`;
                }
            },
            {
                title: "URL",
                data: "url",
                className: "text-truncate",
                render: function(data, type) {
                    const url = sanitise(data);
                    return `<a href="${url}" class="btn btn-link text-start text-decoration-none" target="_blank">${url}</a>`;
                }
            },
            {
                title: "Rules",
                data: "unique_rules",
                className: "text-center text-nowrap",
                render: function(data, type) {
                    let badges = $("<div>");
                    data.forEach(function(rule, idx) {
                        let badge = $(`<span class="badge text-bg-secondary">${rule.name}</span>`)
                        if (idx > 0) { badge.addClass("ms-2") }
                        badges.append(badge);
                    });

                    return badges.html();
                }
            },
            {
                title: "Created At",
                data: "created_at",
                render: function(data, type) {
                    let dateTime = new Date(data);
                    return $(`
                        <span data-bs-trigger="hover focus"
                            data-bs-html="true"
                            data-bs-custom-class="text-center"
                            data-bs-toggle="popover"
                            data-bs-content="${formatStringDate(dateTime, "%a, %D %B, %Y<br>%H:%M:%S")}">
                            ${formatStringDate(dateTime, "%D, %b %Y")}
                        </span>
                    `).popover()[0];
                }
            },
            {
                title: "Notes",
                data: "extra_notes",
                orderable: false,
                className: "text-center",
                render: function(data, type) {
                    if (!data) { return "" }
                    const extraNotes = sanitise(data);
                    return $(`
                        <i class="bi bi-chat-left-text"
                            data-bs-trigger="hover focus"
                            data-bs-toggle="popover"
                            data-bs-title="Extra Notes"
                            data-bs-content="${extraNotes}">
                        </i>
                    `).popover()[0];
                }
            },
            { 
                title: "Active",
                data: "active",
                orderable: false,
                className: "text-center form-switch",
                render: function(data, type) {
                    return `<input type="checkbox" class="sub-toggle-active form-check-input ms-0" ${data ? "checked" : ""} />`
                }
            }
        ]
    );
}

// #endregion

// #region Load Data

$(document).on("selectedServerChange", async function() {
    await loadSubscriptionData();
});

$(subTableId).on("doDataLoad", async function() {
    await loadSubscriptionData();
})

async function loadSubscriptionData() {
    if (!selectedServer) {
        return;
    }

    setTableFilter(subTableId, "server", selectedServer.id);
    await loadTableData(subTableId, `/api/subscriptions/`, "GET");
}

// #endregion

